/*
 * This module is the main code for processing a request received by
 * the http server.
 *
 * The request is specifed by an array of structures that point to strings.
 * The first 3 elements point to the 3 parts of the initial request line
 * (method, URL, protocol-version).  The remaining elements point to
 * header request lines.  This list is always terminated by a zero-length
 * line.  If the caller detected an error while generating the request,
 * the final line will have a negative length.
 */
#include <stdio.h>
#include "tutil.h"
#include "access.h"

typedef struct { int l; char *s; } string;
int http_log_level;		/* Global variable, logging control */
int tlog_put();			/* Logging output routine */
int http_add_response(), http_send_response_header();

int http_execute ( void *ctx,		/* opaque, context for tcp connection */
	char *log_prefix,		/* Tag for log file output */
	string *request,		/* Request to process. */
	tu_stream inbound )		/* Input character stream */
{
    int status, data_bytes, http_send_error(), http_send_document(), uic;
    int http_parse_url(), http_translate_ident(), http_decnet_execute();
    char *service, *node, *ident, *suffix, *arg;
    struct acc_info_struct acc;
    char url_buf[4096], rsphdrtxt[512], munged_ident[300], iobuffer[4096];
    string iobuf;
    struct tu_textbuf rsphdr;
    /*
     * Initialize structure for holding response header lines.
     */
    data_bytes = 0;			/* bytes of data transmitted. */
    rsphdr.l = 0;			/* current length */
    rsphdr.s = rsphdrtxt;		/* data storage */
    rsphdr.size = sizeof(rsphdrtxt);	/* Storage capacity */
    iobuf.l = sizeof(iobuffer);		/* Work area */
    iobuf.s = iobuffer;			/* Work area */
    if ( request[2].l > 0 ) tu_add_text ( &rsphdr, "HTTP/1.0 ", 9 );
    /*
     * Handle error case.
     */
    if ( request[3].l < 0 ) {
	if ( http_log_level > 1 ) tlog_put 
		( "!AZ generating error response!/", log_prefix );
	status = http_send_error ( ctx, &rsphdr, "400 Bad request",
		"Syntax error or malformed request", request );
	return status;
    }
    /*
     * First attempt to translate URL.
     */
    status = http_parse_url ( request[1].s, url_buf, &service, &node, &ident, 
	&arg );
    if ( status ) {
	/*
	 * Check for gateway request (node field is non-blank).
	 */
	if ( *service || *node ) {
	    status = http_send_error ( ctx, &rsphdr, "501 Not implemented",
		"Gateway function not available.", request );
	    return status;
	}
	/*
	 * Local ident, attempt to re-map according to local rules.
	 */
        status = http_translate_ident ( ident, munged_ident, 
			sizeof(munged_ident), &acc );
	if ( !status ) {
	    /*
	     * Error status means ident matched a 'fail' record in the
	     * the rule database.
	     */
	    status = http_send_error ( ctx, &rsphdr, "403 Forbidden",
		"Access to object is _ruled_ out.", request );
	    return status;
	}
	
    } else {
	/*
	 * Parse failure on URL.
	 */
	status = http_send_error ( ctx, &rsphdr, "400 Bad URL",
		"invalid URL specified", request );
	return status;
    }
    /*
     * Take action depending upon method specified in request or punt
     * to HTBIN escape mechanism.
     */
    if ( status == 2 ) {
	/* Rule file redirect */
	tu_strcpy ( iobuffer, "location: " );
	tu_strcpy ( &iobuffer[10], munged_ident );
	status = http_add_response ( &rsphdr, "302, rule file redirect", 0 );
	status = http_add_response ( &rsphdr, iobuffer, 1 );
	status = http_send_response_header ( ctx, &rsphdr );

    } else if ( status == 3 ) {
	/* Exec command. */
	status = http_decnet_execute ( ctx, "HTBIN",
		request, inbound, munged_ident, arg, &rsphdr, 
		&iobuf, &data_bytes );
    } else if ( 0 == tu_strncmp ( request[0].s, "GET", 4 ) ) {
        status = http_send_document ( ctx, request, inbound, ident,
		munged_ident, arg, &rsphdr, &iobuf, &data_bytes, &acc );
    } else if ( 0 == tu_strncmp ( request[0].s, "HEAD", 5 )  ) {
        status = http_send_document ( ctx, request, inbound, ident,
		munged_ident, arg, &rsphdr, &iobuf, &data_bytes, &acc );
	/* status = http_send_error ( ctx, &rsphdr, "501 Not implemented",
		"Method not supported by this server", request ); */
    } else if ( 0 == tu_strncmp ( request[0].s, "POST", 5 )  ) {
        status = http_decnet_execute ( ctx, "POST", request, inbound,
		munged_ident, arg, &rsphdr, &iobuf, &data_bytes );
    } else if ( 0 == tu_strncmp ( request[0].s, "INVCACHE", 9 )  ) {
	int http_invalidate_doc_cache();
	status = http_invalidate_doc_cache();
	status = http_send_error ( ctx, &rsphdr, "200 cache invalidated",
		"All cache entries have been cleared", request );
    } else {
	status = http_send_error ( ctx, &rsphdr, "501 Not implemented",
		"Method not supported by this server", request );
    }
    /*
     * Log data transfer statistics.
     */
    if ( http_log_level > 2 )
    tlog_put ( "!AZ      response stats: hdr = !SL, data = !SL bytes!/",
	log_prefix, rsphdr.l, data_bytes );
    return status;
}
/**************************************************************************/
/* Append text followed by CRLF to response buffer.   If standard flag is true,
 *  append default headers to buffer. If standard flag is 2, use html content.
 */
int http_add_response ( tu_text rsp, char *text, int standard )
{
    int status;
    status = tu_add_text ( rsp, text, 512 );
    status = tu_add_text ( rsp, "\r\n", 3 );
    if ( standard ) {
	status = tu_add_text ( rsp, "MIME-version: 1.0\r\n", 80 );
	    status = tu_add_text ( rsp, (standard == 2) ? 
		"Content-type: text/html\r\n" : 
		"Content-type: text/plain\r\n", 80 );
    }
    return status;
}
/*****************************************************************************/
/*  Generate plain/text document to report error and send to client.
 */
int http_send_error ( void *ctx, tu_text rsphdr, 
	char *status_msg, char *err_text, string *request )
{
    int i, status, ts_tcp_write();
    /*
     * Build and send standard response header.
     */
    status = http_add_response ( rsphdr, status_msg, 1 );
    status = http_send_response_header ( ctx, rsphdr );
    if ( (status&1) == 1 ) {
	/*
	 * Send additional error text as body of message.
	 */
	status = ts_tcp_write ( ctx,"-ERROR-(", 8 );
	status = ts_tcp_write ( ctx, status_msg, 3 );
	status = ts_tcp_write ( ctx, "):  ", 4 );
	status = ts_tcp_write ( ctx, err_text, tu_strlen(err_text) );
	if ( (status&1) == 0 ) return status;
	/*
	 * Send initial request parse result.
	 */
	if ( request[0].l > 0 ) {
	    status = ts_tcp_write ( ctx, "\r\nRequested method: ", 20 );
	    if ( (status&1) == 0 ) return status;
	    status = ts_tcp_write ( ctx, request[0].s, request[0].l );
	}
	if ( request[1].l > 0 ) {
	    status = ts_tcp_write ( ctx, "\r\nRequested URL:    ", 20 );
	    if ( (status&1) == 0 ) return status;
	    status = ts_tcp_write ( ctx, request[1].s, request[1].l );
	}
	if ( request[2].l > 0 ) {
	    status = ts_tcp_write ( ctx, "\r\nHTTP protocol:    ", 20 );
	    if ( (status&1) == 0 ) return status;
	    status = ts_tcp_write ( ctx, request[2].s, request[2].l );
	}
	/*
	 * Send any remaining request lines.
	 */
	status = ts_tcp_write ( ctx, 
	/*               5678901234567890123456789012345678901234567890 */
		"\r\n\r\n-------- additional request headers --------\r\n",50);
	for ( i = 3; (request[i].l > 0) && (status&1); i++ ) {
	    status = ts_tcp_write ( ctx, request[i].s, request[i].l );
	    if ( (status&1) == 0 ) return status;
	    status = ts_tcp_write ( ctx, "\r\n", 2 );
	    if ( (status&1) == 0 ) return status;
	}
    }
    return status;
}
/***************************************************************************/
/* Send contents of response header buffer to client.
 */
int http_send_response_header ( void *ctx, tu_text rsp)
{
    int ts_tcp_write(), status, length;
    char *buffer;

    buffer = rsp->s;
    if ( (rsp->l > 9) && (0 == tu_strncmp(rsp->s,"HTTP/1.0 ", 9)) ) {
	tu_add_text ( rsp, "\r\n", 3 );		/* add null line */
	length = rsp->l;
    } else if ( *buffer == '3' ) {  /* force full re-direct headers */
	/*
	 * Response buffer is old protocol, skip the version field and
	 * truncate after first line.
	 */
	/* buffer = &buffer[9]; */
	for ( length = 0; 
		buffer[length] && (buffer[length] != '\n'); length++);
	if ( buffer[length] == '\n' ) length++;
	buffer = &buffer[length];
	for ( ; buffer[length] && (buffer[length] != '\n'); length++);
	if ( buffer[length] == '\n' ) length++;
    } else {
	/*
	 * Don't send anything.
	 */
	length = 0;
    }
    if ( length > 0 ) status = ts_tcp_write ( ctx, buffer, length );
    else status = 1;
    return status;
}
/***************************************************************************/
int http_parse_url 
	( char *url, 			/* locator to parse */
	char *info,			/* Scratch area for result pts*/
	char **service,			/* Protocol (e.g. http) indicator */
	char **node,			/* Node name. */
	char **ident,			/* File specification. */
	char **arg )			/* Search argument */
	
{
    int i, state;
    char *last_slash, *p, c, arg_c;
    /*
     * Copy contents of url into info area.
     */
    *service = *node = *ident = *arg = last_slash = "";

    for ( state = i = 0; (info[i] = url[i]) != 0; ) {
	c = info[i];
	switch ( state ) {
	    case 0:
		if ( c == ':' ) {
		    info[i] = '\0';	/* terminate string */
		    *service = info;
		    state = 1;
		}
	    case 1:
		if ( c == '/' ) {
		    *ident = last_slash = &info[i];
		    state = 2;
		}
		break;
	    case 2:
		state = 4;
		if ( c == '/' ) {	/* 2 slashes in a row */
		    *node = *ident;
		    state = 3;
		}
		else if ( (c == '#') || (c == '?') ) {
		    arg_c = c;
		    info[i] = '\0';
		    *arg = &info[i+1];
		    state = 5;
		}
		break;
	    case 3:			/* find end of host spec */
		if ( c == '/' ) {
		    state = 4;
		    *ident = last_slash = &info[i];
		    for ( p = *node; p < *ident; p++ ) p[0] = p[1];
		    info[i-1] = '\0';	/* Terminate host string */
		}
		break;
	    case 4:			/* Find end of filename */
		if ( c == '/' ) last_slash = &info[i];
		else if ( (c == '#') || (c == '?') ) {
		    arg_c = c;
		    info[i] = '\0';
		    *arg = &info[i+1];
		    state = 5;
		}
	    case 5:
		break;
        }
	i++;
    }
    /*
     * Insert arg delimiter back into string.
     */
    if ( **arg ) {
	char tmp;
	for ( p = *arg; arg_c; p++ ) { tmp = *p; *p = arg_c; arg_c = tmp; }
	*p = '\0';
    }
    return 1;
}
/*
 * Scan ident and return pointer to suffix porition of filename, including
 * the period.
 * Return "/" if no filename is present and "" if no extension.
 */
char *http_url_suffix ( char *ident )
{
    char *p, *suffix;
    suffix = "";
    for ( p = ident; *p; p++ )
	if ( *p == '.' ) suffix = p;
	else if ( *p == '/' ) suffix = "";
    /*
     * Check if last thing on line is "/".
     */
    if ( (*suffix == '\0') && (p > ident) )
	if ( *--p == '/' ) suffix = p;

    return suffix;
}
/***************************************************************************/ 
