/*
 * This module manages the internal database of pathname (ident) translation
 * and screening information.
 *
 * Note the the define routine is not thread-safe.
 *
 * The suffix string is stored upcase and lookups are case insensitive.
 *
 *   int http_define_ident ( char *ident, int code, char *value );
 *	code:	1 - Add map entry.
 *		2 - Add pass (if match) entry.
 *		3 - Add fail (if match) entry.
 *		4 - Add redirect (if match) entry.
 *		5 - Add exec (htbin) entry.
 *		6 - Add userdir entry
 *		7 - Add protect entry
 *
 *   int http_translate_ident 
 *	   ( char *ident, char *translation, int maxlen, access_info acc );
 *	ret val: 0	Translation failed (matched fail record).
 *		1	Translation successful, ident matched pass entry.
 *		2	Translation successful, return redirect.
 *		3	Translation successful, matched exec entry.
 *
 * Revision history:
 *    25-FEB-1994	Added piggyyback mapping to pass command.
 *     6-APR-1994	Added redirect and generic exec support.
 *     4-MAY-1994	Added support for userdir command.
 *    26-MAY-1994	Added protect option.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "tutil.h"
#include "access.h"
int http_log_level, tlog_put();		/* Global variable, logging level */
/*
 * Define module-wide structures for the database.
 */
struct info_blk {
    struct info_blk *next;		/* succeeding block in chain */
    int code;				/* Entry type. */
    int pat_len;			/* pattern length, not including '*' */
    int val_len;			/* length of value string */
    int wildcard;			/* if nonzero, pattern ends in '*' 
					 * if >1, value ends in wildcard. */
    char *pattern;			/* Substring to find in source */
    char *value;			/* Replacement value for substring */
};
typedef struct info_blk *info_ptr;

static int rule_count = 0;
static info_ptr rule_list, last_rule;

/*********************************************************************/
/*
 * Define new entry for ident map.  Return value is 1 for success, 0 for
 * error.
 */
int http_define_ident ( char *ident, int code, char *value )
{
    info_ptr new;
    int val_len;
    /*
     * Allocate block and its string storage.  Pattern is upcased for
     * case insensitivity.
     */
    new = (info_ptr) malloc ( sizeof(struct info_blk) );
    new->next = (info_ptr) NULL;
    new->code = code;
    new->pat_len = strlen(ident);
    val_len = strlen ( value );
    new->val_len = val_len;
    new->pattern = malloc ( new->pat_len + 1 );
    tu_strupcase ( new->pattern, ident );
    new->value = malloc ( val_len + 1 );
    strcpy ( new->value, value );
    /*
     * See if pattern has wildcard character.
     */
    if ( (new->pat_len > 0) && (new->pattern[new->pat_len-1] == '*') ) {
	new->wildcard = 1;
	--new->pat_len;			/* don't include '*' */
	/*
	 * See if value ends in '*', setting
	 */
	if ( val_len > 0 ) if ( new->value[val_len-1] == '*' ) {
	   /* Trim the last character and change wildcard status. */
	    --new->val_len;
	    new->value[val_len-1] = '\0';
	    new->wildcard = 2;
	}
    }
    else new->wildcard = 0;
    /*
     * Insert at end of list.
     */
    if ( rule_count == 0 ) rule_list = new;
    else last_rule->next = new;
    last_rule = new;
    rule_count++;
    return 1;
}
/***************************************************************************/
/* Process ident against rules files:  Turn
 *
 * Return values:
 *	0	Translation failed (matched fail record).
 *	1	Translation successful, ident matched pass entry.
 *	2	Translation successful, return redirect.
 *	3	Translation successful, matched exec entry.
 *	4	Translation successful, matched userdir entry.
 */
int http_translate_ident ( char *ident, char *translation, int maxlen, 
	access_info acc )
{
    info_ptr blk;
    int i, ident_len, http_translate_userdir(), length, status;
    int http_check_protection();
    char *defprot;
    /*
     * Make initial translation.
     */
    tu_strnzcpy ( translation, ident, maxlen - 1 );
    tu_strupcase ( translation, translation );
    ident_len = tu_strlen ( translation ) + 1;
    acc->uic = 0;
    acc->prot_file = "";		/* No protection file */
    acc->cache_allowed = 1;		/* Caching enabled */
    defprot = "";			/* default protection file */
    /*
     * Process translation against rules.
     */
    for ( blk = rule_list; blk; blk = blk->next ) {
	int cmp;
	/*
	 * Either compare entire string or only up to wildcard position.
	 */
	cmp = tu_strncmp ( translation, blk->pattern, 
		blk->wildcard ? blk->pat_len : ident_len );
	if ( http_log_level > 10 ) tlog_put ( 
		"Compared  !SL '!AZ' with '!AZ', result: !SL!/", blk->code,
		translation, blk->pattern, cmp );

	if ( cmp == 0 ) {
	    /*
	     * We matched pattern, Take action based upon code.
	     */
	    switch ( blk->code ) {
		case 2:			/* Pass on ident match, possibly */
					/* mapping result. */
		    if ( blk->val_len <= 0 ) return 1;

		case 1:		/* map */
		case 4:		/* redirect */
		    /* Replace either all of translation or part with
		     *	new value
		     */
		    if ( blk->wildcard > 1 ) {
			/* 
			 * replace matching portion only.
			 */
			int replace_len, delta;
			delta = blk->val_len - blk->pat_len;
			if ( delta > 0 ) {
			    /* New value longer than current */
			    if ( ident_len + 1 + delta > maxlen ) return 0;
			    for ( i = ident_len; i >= blk->pat_len; --i )
				translation[i+delta] = translation[i];
			} else {
			    /*
			     * Easy case, new is smaller than old.
			     */
			    for ( i = blk->pat_len; i <= ident_len; i++ )
				translation[i+delta] = translation[i];
			
			}
			ident_len = ident_len + delta;
		        tu_strncpy ( translation, blk->value, blk->val_len );
		    } else {
			/* Replace entire thing. */
			tu_strnzcpy ( translation, blk->value, maxlen - 1 );
		        ident_len = tu_strlen ( translation ) + 1;
		    }
		    if ( http_log_level > 1 ) tlog_put ( 
			"ident_map: '!AZ' -> '!AZ'!/", ident, translation );
		    /*
		     * Either return success or try next rule.
		     * Don't upcase redirect values.
		     */
		    if ( blk->code == 4 ) return 2;	/* redirect */
		    tu_strupcase ( translation, translation );
		    if ( blk->code == 2 ) return 1;	/* pass */
		    break;
		case 3:
		    return 0;		/* fail on ident match */
		case 5:		/* exec */
		    /*
		     * Exec directive, append * pattern * value to translation.
		     */
		    i = tu_strlen ( translation );
		    translation[i++] = '*';		/* first delimiter */
		    tu_strnzcpy ( &translation[i], blk->pattern,
				blk->pat_len );
		    i += blk->pat_len;
		    translation[i++] = '*';
		    length = maxlen - i - 1;
		    if ( length > blk->val_len ) length = blk->val_len;
		    tu_strnzcpy ( &translation[i], blk->value,  length );
		    return 3;
		case 6:
		    /*
		     * userdir directive, replace ~username/ with login
		     * directory + translation + /
		     */
		    status = http_translate_userdir
			(blk, translation, maxlen, &acc->uic);
		    if ( http_log_level > 1 ) tlog_put ( 
			"ident_map(user): '!AZ' -> '!AZ'!/", ident, translation );
		    return status;
		case 7:
		    /*
		     * Access to the current ident is protected.  Return
		     * failure if protection check fails.
		     */
		    acc->prot_file = blk->value;
		    if ( *acc->prot_file == '\0' ) acc->prot_file = defprot;
/*		    status = http_check_protection 
				( blk->value, translation, acc );
		    if ( status == 0 ) return status; */
		    break;
		case 8:		/* defprot rule */
		    defprot = blk->value;
	    }
	}
    }
    return 0;		/* default to failure status */
}

/***************************************************************************/
/* Convert /~username/.... specification to /device/home-dir/... according
 * to information in the user authorization file.  The sysuaf file may need to
 * be modified to granted read access.
 *
 * Return values:
 *	0	Translation failed (username not found)
 *	4	Translation successful.
 */
int http_translate_userdir ( info_ptr blk, char *translation, int maxlen,
	int *uic )
{
    int i, user_len, login_len, delta, status, length;
    int http_user_information();
    char login[128];
    /*
     * Parse username out of translation.
     */
    for ( i = blk->pat_len; (i < maxlen) && translation[i] &&
	  (translation[i] != '/'); i++ );
    user_len = i - blk->pat_len;
    /*
     * Fetch information abour user.  Block other activity and set privilege.
     */
    status = http_user_information 
	( &translation[blk->pat_len], user_len, uic, login, sizeof(login) );
    if ( (status&1) == 0 ) return 0;
    /*
     * Calculate length of login directory + userdir path and make gap
     * in translation string to accomodate it.
     */
    login_len = tu_strlen ( login );
    user_len += blk->pat_len;
    delta = login_len + blk->val_len - user_len;

    length = tu_strlen ( translation );
    if ( delta > 0 ) {
	/* New value longer than current */
	if ( length+delta > maxlen ) delta = 0;
	for ( i = length; i > 0; --i )
		translation[i+delta] = translation[i];
    } else {
	/* New value smaller than /~username. */
	for ( i = user_len; i <= length; i++ ) 
		translation[i+delta] = translation[i];
    }
    /*
     * Load users device, directory and subdirectory into gap.
     */
    tu_strncpy ( translation, login, login_len );
    tu_strncpy ( &translation[login_len], blk->value, blk->val_len );
    /*
     * Do additional fixup of user portion.  Search for "/../" and disallow
     * translation.
     */
    tu_strupcase ( &translation[login_len], &translation[login_len] );
    for ( i = login_len; translation[i]; i++ ) if ( translation[i] == '/' ) {
	if ( 0 == strncmp ( &translation[i], "/../", 4 ) ) return 0;
    }

    return 4;
}
