/*
 * This module provides output to an application log file.  It formats
 * data via SYS$FAO to facilitate putting timestamps in the control strings.
 *
 *  int tlog_init ( char *log_file_name );
 *  int tlog_put ( char *ctlstr, ... );
 *  int tlog_flush;
 */
#include <stdio.h>
#include <stdarg.h>
#include <descrip.h>
#include <ssdef.h>
#include <string.h>
#include "pthread_np.h"

static FILE *log_file;
static int log_valid = 0;
static struct dsc$descriptor control;	/* control string */
static struct dsc$descriptor output;	/* control string */
static char log_file_spec[256];
static pthread_mutex_t log_write;
static int flush_ef;
static long flush_interval[2];
static int flush_enable;
int LIB$GET_EF(), SYS$CLREF(), SYS$SETIMR();

int tlog_init ( char *log_file_name )
{
    if ( log_valid ) return 3;
    log_file = fopen ( log_file_name, "a",
	"dna=.log",
	"alq=100", "deq=600", /* "bls=8192", */
	"rfm=var", "mrs=300",
	"ctx=rec", "shr=upd");
    strcpy ( log_file_spec, log_file_name );
    if ( log_file ) log_valid = 1;
    /*if ( log_file ) { log_valid = 1; fgetname ( log_file, log_file_spec );}*/
    if ( !log_valid ) perror ( "Can't open log file" );
    LIB$GET_EF ( &flush_ef );
    SYS$CLREF ( flush_ef );
    flush_enable = 0;
    flush_interval[0] = -600000000; flush_interval[1] = -1;
    SYS$SETIMR ( flush_ef, &flush_interval, 0, 0, 0 );

    /*
     * initialize static fields in descriptors.
     */
    control.dsc$b_dtype = DSC$K_DTYPE_T;
    control.dsc$b_class = DSC$K_CLASS_S;
    output.dsc$b_dtype = DSC$K_DTYPE_T;
    output.dsc$b_class = DSC$K_CLASS_S;
    /*
     * Initialize mutexes and conditions.
     */
    pthread_mutex_init ( &log_write, pthread_mutexattr_default );

    return log_valid;
}

int tlog_put ( char *ctlstr, ... )
{
    va_list param_list;
#ifdef __ALPHA
    unsigned int __VA_COUNT_BUILTIN(void), arg_count;
    int i, fao_param[32];
#endif
    int status, SYS$FAOL(), length, clen;
    char outbuf[1000];
    va_start(param_list,ctlstr);
    /*
     * See if log file available and synchronize
     */
    if ( !log_valid ) return 0;
    pthread_mutex_lock ( &log_write );
    for ( clen = 0; ctlstr[clen]; clen++ );
    control.dsc$w_length = clen;
    control.dsc$a_pointer = ctlstr;
    output.dsc$w_length = sizeof(outbuf);
    output.dsc$a_pointer = outbuf;

    length = 0;
    LOCK_C_RTL
#ifdef __ALPHA
    arg_count = __VA_COUNT_BUILTIN() - 1;
    for ( i = 0; i < arg_count; i++ ) fao_param[i] = va_arg(param_list,int);
    status = SYS$FAOL ( &control, &length, &output, fao_param );
#else
    status = SYS$FAOL ( &control, &length, &output, param_list );
#endif
    if ( (status&1) == 1 ) {
	status = fwrite(outbuf, length, 1, log_file);
	if ( status != 1 ) {
	    char errtxt[32];
	    sprintf(errtxt, "log file write error (length=%d)", length );
	    perror ( errtxt );
	}
    } else {
	fprintf(log_file,"FAO error: %d '%s'\n", status, ctlstr );
    }
    UNLOCK_C_RTL
    flush_enable = 1;
    pthread_mutex_unlock ( &log_write );
    return status;
}


int tlog_flush()
{
    if ( log_valid ) {
	int status, ef_mask, SYS$READEF();
	pthread_mutex_lock ( &log_write );
	if ( flush_enable ) {
	    flush_enable = 0;
	    if ( SYS$READEF(flush_ef,&ef_mask) == SS$_WASSET ) {
		SYS$CLREF(flush_ef);
		SYS$SETIMR ( flush_ef, flush_interval, 0, 0, 0 );
		log_file = freopen ( log_file_spec, "a", log_file,
			"dna=.log",
			"alq=100", "deq=600", /* "bls=8192", */
			"rfm=var", "mrs=300",
			"ctx=rec", "shr=upd");
		if ( !log_file ) {
		    log_valid = 0;
		    printf("Error on re-open of log file\n");
		}
	    }
	} else status = 0;
	pthread_mutex_unlock ( &log_write );
	return status;
    }else return 0;
}
